#
# Copyright 2009 by Object Craft P/L, Melbourne, Australia.
#
# LICENCE - see LICENCE file distributed with this software for details.
#
# $Id: test_form.py 15579 2010-03-15 01:18:36Z bgg $

import os
import unittest
import difflib

from albatross import SimpleContext, ApplicationError

from albatross.ext.form import *


class Model:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class FormTestCase(unittest.TestCase):

    def setUp(self):
        self.name_field = TextField('Name', 'name', required=True)
        self.age_field = IntegerField('Age', 'age')
        self.weight_field = FloatField('Weight', 'weight')
        fields = (
            self.name_field,
            self.age_field,
            self.weight_field
        )
        tags = (Fieldset(fields), )
        self.form = FieldsetForm('Test Form', tags)

    def _run_form(self, form, opts=FormOptions()):
        """
        pretend we've just had data lodged back into the form
        from the browser, ie, convert all values to strings.
        """
        ctx = SimpleContext('.')
        ctx.push_content_trap()
        form.run(ctx, 'test', opts)
        content = ctx.pop_content_trap()
        return content

    def test_form_validation(self):
        model = Model(name='Fred', age=18, weight=80.5)
        self.form.load(model)
        self._run_form(self.form)
        self.form.validate()    # Initial values should validate
        self.name_field.value = ''
        self.age_field.value = 'twelve'
        self.weight_field.value = 'pants'
        # New values should not validate
        self.assertRaises(FormValidationError, self.form.validate)
        self.failIf(self.name_field
                    not in self.form.validation_errors,
                    'Expected validation to fail for name field')
        self.failIf(self.age_field
                    not in self.form.validation_errors,
                    'Expected validation to fail for age field')
        self.failIf(self.weight_field
                    not in self.form.validation_errors,
                    'Expected validation to fail for weight field')

    def test_form_merge(self):
        model = Model()
        self.name_field.set_value('Bob', is_display_value=True)
        self.age_field.set_value('42', is_display_value=True)
        self.weight_field.set_value('82.5', is_display_value=True)

        self.form.validate()    # New values should validate.
        self.form.merge(model)
        self.failIf(model.name != 'Bob',
                    'Unexpected value %r for name' % model.name)
        self.failIf(model.age != 42,
                    'Unexpected value %r for age' % model.age)
        self.failIf(model.weight != 82.5,
                    'Unexpected value %r for weight' % model.weight)

    def test_clear(self):
        model = Model(name='Fred', age=18, weight=80.5)
        self.form.load(model)
        self._run_form(self.form)
        self.form.validate()    # Initial values should validate
        self.form.clear()
        self.failIf(self.name_field.value,
                    'Unexpected value %r for name' % self.name_field.value)
        self.failIf(self.age_field.value,
                    'Unexpected value %r for age' % self.age_field.value)
        self.failIf(self.weight_field.value,
                    'Unexpected value %r for weight' % self.weight_field.value)

    def test_text_field_none(self):
        model = Model(name=None, age=27, weight=65.7)
        self.form.load(model)
        # Ensure 'None' is converted to an empty string in the form.
        self.failIf(self.name_field.get_value() != '',
                    'Unexpected value %r for name_field.value'
                    % self.name_field.value)

    def test_int_field(self):
        field_name = 'Weight'
        fields = (
            IntegerField(field_name, 'weight'),
        )
        form = Form('Form', fields)
        w = 80.7
        model = Model(weight=w)
        form.load(model)
        weight = form.get_field(field_name)
        self.failIf(weight.get_value() != w,
                    'failed to get int field, exp %r, got %r' % (w, weight.value))
        self._run_form(self.form)
        self.failIf(weight.get_value() != w,
                    'failed to get int field, exp %r, got %r' % (w, weight.value))        

    def expect_html(self, result, expect):
        self.assertEqual(result, expect,
                         'expected "%s", got "%s"\n%s' % (expect, result,
                                                          ''.join(difflib.context_diff(expect, result))))

    def load_html(self, filename):
        f = open(os.path.join(os.path.dirname(__file__), 'template', filename))
        try:
            return f.read()
        finally:
            f.close()

    def test_fieldset_html(self):
        # Slightly abridged list of all the countries in the world.
        country_names = (
            'Australia',
            'Belgium',
            'Cuba',
            'Greenland',
            'Madagascar',
            'Netherlands',
            'Switzerland',
            'Uzbekistan',
            'Zimbabwe'
        )
        country_menu = [e  for e in enumerate(country_names)]
        fields = (
            TextField('Name', 'name', required=True),
            IntegerField('Integer', 'an_int'),
            FloatField('Float', 'a_float'),
            SelectField('Country', country_menu, 'country'),
            PasswordField('Password', 'password', required=True),
            # with a checkbox, in the generated HTML value="on"
            # but note that "checked" is not set
            Checkbox('Active', 'active'),
        )
        buttons = Buttons((
            Button('Save', 'save'),
            Button('Cancel', 'cancel'),
        ))
        elements = (Fieldset(fields),)
        test_form = FieldsetForm('User details',
                                 elements,
                                 buttons=buttons)
        content = self._run_form(test_form)
        name_field = test_form.get_field('Name')
        self.expect_html(content,
                         self.load_html('fieldset.html'))

    def test_checkbox_on_off(self):
        fields = (
            Checkbox('Checkbox on', 'cb1_value', value=True),
            Checkbox('Checkbox off', 'cb2_value', value=False),
        )
        form = Form('Form', fields)
        content = self._run_form(form)
        checkbox_on = form.get_field('Checkbox on')
        checkbox_off = form.get_field('Checkbox off')
        self.failIf(not checkbox_on.get_value(),
                    'checkbox_on should be on, got %s' % checkbox_on.get_value())
        self.failIf(checkbox_off.get_value(),
                    'checkbox_off should be off, got %s' % checkbox_off.get_value())

    def test_checkbox_on_off_static(self):
        fields = (
            Checkbox('Checkbox on', 'cb1_value', value=True),
            Checkbox('Checkbox off', 'cb2_value', value=False),
        )
        form = Form('Form', fields)
        content = self._run_form(form, opts=FormOptions(static=True))
        self.expect_html(content,
                         self.load_html('checkbox-static.html'))

    def test_radio(self):
        radio_menu = (
            (1, 'one'),
            (2, 'two'),
            (3, 'three'),
        )
        fields = (
            RadioField('Radio', radio_menu, 'radio_value', value=1),
        )
        form = Form('Form', fields)
        content = self._run_form(form)
        self.expect_html(content,
                         self.load_html('radiofield.html'))
        content = self._run_form(form, opts=FormOptions(static=True))
        self.expect_html(content,
                         self.load_html('radiofield-static.html'))

    def test_textarea(self):
        fields = (
            Textarea('Textarea', 'text'),
        )
        form = Form('Form', fields)
        content = self._run_form(form)
        self.expect_html(content,
                         self.load_html('textarea.html'))

    def test_select(self):
        numbers = ['zero', 'uno', 'due', 'tre']
        number_menu = [e  for e in enumerate(numbers)]
        fields = (
            SelectField('Number', number_menu, 'number'),
        )
        form = Form('Form', fields)
        content = self._run_form(form)
        number_field = form.get_field('Number')
        self.failIf(number_field.get_value() != 0,
                    'select should be 0; got %r' % number_field.get_value())

    def test_get_field(self):
        field_name = 'Integer'
        fields = (
            IntegerField(field_name, 'an_int'),
        )
        form = Form('Form', fields)
        f = form.get_field(field_name)
        self.failIf(f is None,
                    'could not find "%s" field in form %s' % (field_name, form))

    def test_get_field_nested(self):
        field_name = 'Integer'
        elements = (
            Table(None, (
                Row((
                    Col((IntegerField(field_name, 'an_int'), )),
                )),
            )),
        )
        form = Form('Form', elements)
        f = form.get_field(field_name)
        self.failIf(f is None,
                    'could not find "%s" field in nested form %s' % (field_name, form))

    def test_field_set_value(self):
        value = 17.0
        float_field = FloatField('Value', None, value=value)
        self.failIf(float_field.value != value,
                    'value not set, exp %s; got %s' % (value, float_field.value))

    def test_form_value_clear(self):
        value = 17.0
        float_field = FloatField('Value', None, value=value)
        self.failIf(float_field.value != value,
                    'value not set, exp %s; got %s' % (value, float_field.value))
        form = FieldsetForm('Value Form', (float_field, ))
        self.failIf(float_field.value != value,
                    'value changed adding to form, exp %s; got %s' % (value, float_field.value))
        form.clear()
        self.failIf(float_field.value is not None,
                    'value not reset, exp %s; got %s' % (None, float_field.value))
        
    def test_form_static_value_clear(self):
        value = 17.0
        float_field = FloatField('Value', None, value=value, static=True)
        self.failIf(float_field.value != value,
                    'value not set, exp %s; got %s' % (value, float_field.value))
        form = FieldsetForm('Value Form', (float_field, ))
        self.failIf(float_field.value != value,
                    'value changed adding to form, exp %s; got %s' % (value, float_field.value))
        form.clear()
        self.failIf(float_field.value != value,
                    'value reset, exp %s; got %s' % (value, float_field.value))

    def test_display_value(self):
        fields = (
            IntegerField('Age', 'age'),
            TextField('Name', 'name'),
            FloatField('Weight', 'weight')
        )
        form = Form('Form', fields)
        model = Model(name='Anne Hathaway', weight=67.3, age=27)
        form.load(model)
        ctx = SimpleContext('.')

        name = form.get_field('Name').get_display_value(ctx, form)
        self.failIf(model.name != name,
                    'name error, got %r; exp %r' % (name, model.name))
        age = form.get_field('Age').get_display_value(ctx, form)
        self.failIf(str(model.age) != age,
                    'age error, got %r; exp %r' % (age, str(model.age)))
        weight = form.get_field('Weight').get_display_value(ctx, form)
        self.failIf(str(model.weight) != weight,
                    'weight error, got %r; exp %r' % (weight, str(model.weight)))

    def test_simple_table(self):
        headers = Row((
            HeaderCol((Label('Name'), )),
        ))
        elements = (
            Table(headers, (
                Row((
                    Col((TextField('Name', 'name'), )),
                )),
            )),
        )
        form = Form('Form', elements)
        model = Model(name='Anne Hathaway')
        form.load(model)
        
        content = self._run_form(form)
        self.expect_html(content,
                         self.load_html('simple_table.html'))

    def test_validate_ok(self):
        fields = (
            IntegerField('Age', 'age'),
            TextField('Name', 'name'),
            FloatField('Weight', 'weight')
        )
        form = Form('Form', fields)
        model = Model(name='Anne Hathaway', weight=67.3, age=27)
        form.load(model)
        ctx = SimpleContext('.')
        opts = FormOptions()
        form.run(ctx, 'form', opts)
        # shouldn't be any problems here
        form.validate()

    def test_validate_fail(self):
        fields = (
            IntegerField('Age', 'age'),
            TextField('Name', 'name'),
            FloatField('Weight', 'weight')
        )
        form = Form('Form', fields)
        model = Model(name='Anne Hathaway', weight=67.3, age=27)
        form.load(model)
        ctx = SimpleContext('.')
        opts = FormOptions()
        form.run(ctx, 'form', opts)
        
        age_field = form.get_field('Age')
        age_field.set_value('twenty seven')
        self.assertRaises(FormValidationError,
                          form.validate)


class TableRow(IteratorTableRow):

    def __init__(self, model):
        cols = (
            Col((TextField('Name', 'name'), )),
            Col((IntegerField('Age', 'age'), )),
            Col((FloatField('Weight', 'weight'), )),
        )
        IteratorTableRow.__init__(self, cols)
        
        self.load(model)


class TableForm(FieldsetForm):

    def __init__(self, model_list):
        headers = Row((
            HeaderCol((Label('Name'), )),
            HeaderCol((Label('Age'), )),
            HeaderCol((Label('Weight'), )),
        ))

        self.table = IteratorTable('table', headers, TableRow, model_list)
        FieldsetForm.__init__(self, 'Model table', (self.table, ))


class TableFormTestCase(unittest.TestCase):

    def setUp(self):
        self.models = (
            Model(name='Anne Hathaway', age=27, weight=68.03),
            Model(name='Kate Moss', age=35, weight=47.62),
        )
        self.table_form = TableForm(self.models)

    def test_add_invalid_row(self):
        self.assertRaises(FormError,
                          self.table_form.table.append_row, 
                          Row((Col((Label('xxx'),)),)))

    def test_validate_ok(self):
        ctx = SimpleContext('.')
        opts = FormOptions()
        self.table_form.run(ctx, 'form', opts)
        # no errors expected
        self.table_form.validate()

    def test_validate_fail(self):
        ctx = SimpleContext('.')
        opts = FormOptions()
        self.table_form.run(ctx, 'form', opts)
        # no errors expected
        self.failIf(not self.table_form.table.children,
                    'table_form has no children?!')
        anne_form = self.table_form.table.children[0]
        age_field = anne_form.get_field('Age')
        age_field.set_value('twenty seven')
        self.assertRaises(FormValidationError,
                          self.table_form.validate)

    def test_append(self):
        model = Model(name='Naomi Campbell', age=39, weight=51.2)
        initial_len = len(self.table_form.table.children)
        self.table_form.table.append(model)
        self.failIf(len(self.table_form.table.children) != initial_len + 1,
                    'len should be %s + 1; got %s' % (initial_len,
                                                      len(self.table_form.table.children)))

    def test_remove(self):
        model = self.models[1]
        initial_len = len(self.table_form.table.children)
        for child in self.table_form.table.children:
            name_field = child.get_field('Name')
            if name_field.get_value() == model.name:
                self.table_form.table.remove(child)
        self.failIf(len(self.table_form.table.children) != initial_len - 1,
                    'len should be %s - 1; got %s' % (initial_len,
                                                      len(self.table_form.table.children)))

    def test_remove_nx(self):
        model = Model(name='Naomi Campbell', age=39, weight=51.2)
        self.assertRaises(FormError,
                          self.table_form.table.remove, model)

    def test_set_disabled(self):
        # check none are initally disabled
        for child in self.table_form.table.children:
            self.failIf(child.disabled,
                        'child %r is disabled?!?' % child)
        self.table_form.set_disabled(True)
        for child in self.table_form.table.children:
            self.failIf(not child.disabled,
                        'child %r is not disabled?!?' % child)


class PagedListTestCase(unittest.TestCase):
    def test_empty_no_pagesize(self):
        paged_list = PagedList([])

        self.failIf(paged_list.pagesize,
                    'pagesize should not be set, got %r' % paged_list.pagesize)
        self.failIf(paged_list.items_on_page(1),
                    'page 1 should be empty, got %r' % paged_list.items_on_page(1))
        self.failIf(paged_list.page_count() != 1,
                    'empty list fits on 1 page')
        self.failIf(paged_list.has_next(1),
                    'empty list never has_next')
        self.failIf(paged_list.has_prev(1),
                    'empty list never has_prev')

    def test_empty_pagesize(self):
        paged_list = PagedList([], pagesize=10)

        self.failIf(paged_list.items_on_page(1),
                    'page 1 should be empty, got %r' % paged_list.items_on_page(1))
        self.failIf(paged_list.page_count() != 1,
                    'empty list should fit on 1 page, got %r' % paged_list.page_count() )
        self.failIf(paged_list.has_next(1),
                    'empty list never has_next')
        self.failIf(paged_list.has_prev(1),
                    'empty list never has_prev')

    def test_page_count(self):
        pagesize = 5
        for n in range(12):
            paged_list = PagedList(range(n), pagesize=pagesize)
            if n <= 1:
                exp_page_count = 1
            else:
                exp_page_count = (n - 1) / pagesize + 1
            self.failIf(paged_list.page_count() != exp_page_count,
                        'page_count error (len %d, pagesize %d): got %d; exp %d' %
                        (n, paged_list.pagesize, paged_list.page_count(), exp_page_count))

    def test_has_next(self):
        n = 12
        pagesize = 5
        paged_list = PagedList(range(n), pagesize=pagesize)
        for page in range(5):
            exp_has_next = (page < (n - 1) / pagesize + 1)
            self.failIf(exp_has_next != paged_list.has_next(page),
                        'has_next(%d) (len %d, pagesize %d) %r; exp %r' %
                        (page, n, pagesize, paged_list.has_next(page), exp_has_next))

    def test_has_prev(self):
        n = 12
        pagesize = 5
        paged_list = PagedList(range(n), pagesize=pagesize)
        for page in range(5):
            exp_has_prev = (page > 1)
            self.failIf(exp_has_prev != paged_list.has_prev(page),
                        'has_prev(%d) (len %d, pagesize %d) %r; exp %r' %
                        (page, n, pagesize, paged_list.has_next(page), exp_has_prev))

    def test_start_index_of_page(self):
        n = 12
        pagesize = 5
        paged_list = PagedList(range(n), pagesize=pagesize)
        for page in range(-1, 5):
            if page < 1:
                exp_start_index = 0
            elif page >= paged_list.page_count():
                exp_start_index = (paged_list.page_count() - 1) * pagesize
            else:
                exp_start_index = (page - 1) * pagesize
            self.failIf(paged_list.start_index_of_page(page) != exp_start_index,
                        'start_index_of_page(%d) (len %d, pagesize %d) %r; exp %r' %
                        (page, n, pagesize, paged_list.start_index_of_page(page), exp_start_index))


class FormOptionsTestCase(unittest.TestCase):

    def test_show_errors(self):
        opts = FormOptions()
        self.failIf(opts.show_errors,
                    'option show_errors should be False')
        opts = FormOptions(show_errors=True)
        self.failIf(not opts.show_errors,
                    'option show_errors should be True')

    def test_static(self):
        opts = FormOptions()
        self.failIf(opts.static,
                    'option static should be False')
        opts = FormOptions(static=True)
        self.failIf(not opts.static,
                    'option static should be True')

    def test_pagesize(self):
        opts = FormOptions()
        self.failIf(opts.pagesize,
                    'pagesize should be 0 or None; got %r' % opts.pagesize)
        pagesize = 15
        opts = FormOptions(pagesize=pagesize)
        self.failIf(opts.pagesize != pagesize,
                    'pagesize should be %r; got %r' % (pagesize, opts.pagesize))
        
        self.assertRaises(ApplicationError,
                          FormOptions, (), pagesize='abc')


if __name__ == '__main__':
    unittest.main()
